'''
$Id: GenericUserFolder.py,v 1.13 1999/11/17 07:23:37 zen Exp $
'''

import Globals, App.Undo
from Globals import HTMLFile, MessageDialog
from string import split,join
from Acquisition import Implicit
from OFS.SimpleItem import Item
from OFS.Document import Document
from base64 import decodestring, encodestring
from urllib import quote, unquote
from ImageFile import ImageFile
from AccessControl.Role import RoleManager
from AccessControl.User \
    import BasicUser, BasicUserFolder, _remote_user_mode,nobody
from AccessControl.PermissionRole import _what_not_even_god_should_do
from OFS.Folder import Folder
from time import time
import types
import sha
import whrandom
import Globals
import threading
from zLOG import LOG

from User import User # Our user object

# Moved to thread shared caches rather than _v_ variables for everything,
# as this allows us to flush our caches
_hashmap = {} # _hashmap[gufid][hash] == (username,lastseentime)
_expires = {}
_usermapcache = {} 
_usernamecache = {}
_userlistcache = {}
_lock = {}
_locklock = threading.RLock() # A lock for the _lock map for catch22 
			      # concurrency paranoia.

def _clen(obj = None):
    'GenericUserFolder.cachestats helper function'
    if obj is None:
	return 0
    else:
	return len(obj)

def _debug(msg): pass
#def _debug(msg): LOG('GUF',0,msg)

class GenericUserFolder(Folder,BasicUserFolder):
    '''A Generic User authentication userlist'''

    meta_type='Generic User Folder'
    id       ='acl_users'
    title    ='Generic User Folder'
    icon     ='misc_/GenericUserFolder/Folder_icon'

    isPrincipiaFolderish=1
    isAUserFolder=1
    isAGenericUserFolder=1

    __ac_permissions__=(
	('View management screens',
	    ['cachestats','flushcache','manage_cache'],['Manager',]),
	('Manage users', 
	    ['getUserNames','getUser','getUsers','manage_listUsers',
	     'manage_userInfo','manage_main'],['Manager',]),
 	('Manage properties',
	    ['manage_edit','manage_properties'],['Manager',]),
	('Can Login and Logout',
	    ['login','logout'],['Anonymous','Manager']))

    manage_options=(
	{'label':'Contents',	'action':'manage_main'},
	{'label':'User List',	'action':'manage_listUsers'},
	{'label':'Cache',	'action':'manage_cache'},
	{'label':'User Info',	'action':'manage_userInfo'},
	{'label':'Properties',	'action':'manage_properties'},
	{'label':'Security',	'action':'manage_access'},
	{'label':'Undo',	'action':'manage_UndoForm'}
    )

    manage_properties=HTMLFile('manage_properties', globals())
    manage_listUsers=HTMLFile('manage_listUsers',globals())
    manage_userInfo=HTMLFile('manage_userInfo',globals())
    manage_cache=HTMLFile('manage_cache',globals())

    _gufid = None

    def __init__(self, cookie_mode=0, REQUEST=None):
        self.cookie_mode=cookie_mode
	self.userlist_expires=900
	self.user_expires=900
	self.cookie_expires=900

	self.roles = ['Anonymous',]

	# Do we already have a unique, unchanging ID per object?
	self._gufid = encodestring(sha.new('%f%f%s%s'  \
		% (time(),whrandom.random(),str(self),str(REQUEST))).digest())

	self._init_caches()
    
    def __setstate__(self,state):
	"Initialize on reload from ZODB"
	Folder.__setstate__(self,state)
	self._init_caches()

    def _init_caches(self):
	gufid = self._gufid

	_locklock.acquire()
	try:
	    if not _lock.has_key(gufid):
		_lock[gufid] = threading.RLock()
		_expires[gufid] = 0
		_usermapcache[gufid] = {}
		_usernamecache[gufid] = None
		_userlistcache[gufid] = None
		_hashmap[gufid] = {}
	finally:
	    _locklock.release()

    def _expire(self,flush=0):
	'''Expire old objects. We do lots of caching, as the methods we
	   use to populate our userfolder could be very, very expensive.
	   Especially for those of us integrating with badly designed
	   databases :-)'''

	gufid = self._gufid
	_lock[gufid].acquire()
	try:
	    if flush or time() > _expires[gufid] + self.user_expires: 
		_expires[gufid] = time()

		_usernamecache[gufid] = None
		_userlistcache[gufid] = None

		if flush:
		    _usermapcache[gufid] = {}

		else:
		    # Create a thread to do this if it blocks for too long
		    usermap = _usermapcache[gufid]
		    for k in usermap.keys():
			if (usermap[k] is None) or usermap[k]._expire():
			    del usermap[k]

		    hashmap = _hashmap[gufid]
		    for k in hashmap.keys():
			if time() > hashmap[k][1] + self.cookie_expires:
			    del hashmap[k]
	finally:
	    _lock[gufid].release()

    def cachestats(self):
	'Return meaningless statistics'
	gufid = self._gufid
	_lock[gufid].acquire()
	try:
	    self._expire()
	    ulexpire = str(_expires[gufid] + self.userlist_expires - time())

	    return (\
		{'text':	'Items in Authorization Token cache',
		 'value':	_clen(_hashmap[gufid])},
		{'text':	'Cached User objects',
		 'value':	_clen(_usermapcache[gufid])},
		{'text':	'Items in getUserNames cache',
		 'value':	_clen(_usernamecache[gufid])},
		{'text':	'Items in getUsers cache',
		 'value':	_clen(_userlistcache[gufid])},
		{'text':	'Seconds until userlist expiry',
		 'value':	str(ulexpire)},
		{'text':'Seconds for unseen Authentication token to expire',
		 'value':	self.cookie_expires},
		{'text':	'Seconds before a cached user is expired',
		 'value':	self.user_expires})

	finally:
	    _lock[gufid].release()
    
    def flushcache(self,REQUEST):
	'Flush the caches'

	gufid = self._gufid
	_lock[gufid].acquire()
	try:
	    self._expire(1)
	    if REQUEST.has_key('flushcookies') and REQUEST.flushcookies:
		_hashmap[gufid] = {}
		LOG('GUF',0,"Authentication tokens flushed by %s" \
		    % REQUEST.AUTHENTICATED_USER.getUserName())
	finally:
	    _lock[gufid].release()

        if REQUEST:
            return MessageDialog(
                   title  ='GenericUserFolder Caches Flushed',
                   message='GenericUserFolder caches have been flushed',
                   action = './manage_cache')
	
    def getUserNames(self):
        '''Returns a list of user names or [] if no users exist'''

	gufid = self._gufid
	_lock[gufid].acquire()
	try:
	    self._expire()
	    if _usernamecache[gufid] is not None:
		return _usernamecache[gufid]

	    if hasattr(self,'userList'):
		data = self.userList(self,self.REQUEST)
		if type(data) == types.StringType:
		    data = split(data)
		data = list(data)

		if len(data) > 0 and type(data[0]) != types.StringType:
		    if hasattr(data[0],'username'):
			for i in range(0,len(data)):
			    data[i] = data[i].username
		    elif hasattr(data[0],'uname'):
			for i in range(0,len(data)):
			    data[i] = data[i].uname
		    
		    if type(data[0]) != types.StringType:
			raise 'Unsupported Type',\
			    'userList method returned an unsupported type'

	    else:
		_usernamecache[gufid] = []
		return _usernamecache[gufid]

	    data.sort()
	    _usernamecache[gufid] = data
	    return data

	finally:
	    _lock[gufid].release()


    def getUsers(self):
        """Return a list of user objects or [] if no users exist"""

	gufid = self._gufid
	_lock[gufid].acquire()
	try:

	    # Expire caches if necessary
	    self._expire()

	    # If we have a valid cached version, return it
	    if _userlistcache[gufid] is not None:
		return _userlistcache[gufid]

	    data = self.getUserNames() # Note this is sorted already

	    userlist   = []
	    usermap    = _usermapcache[gufid]

	    for name in data:
		if usermap.has_key(name):
		    user = usermap[name]
		else:
		    user = User(name,self)
		    usermap[name] = user
		user = user.__of__(self)
		userlist.append(user)

	    _userlistcache[gufid] = userlist

	    return userlist
	
	finally:
	    _lock[gufid].release()


    def getUser(self, name):
        """Return the named user object or None if no such user exists"""

	gufid = self._gufid
	_lock[gufid].acquire()

	try:
	    self._expire()

	    cache = _usermapcache[gufid]

	    # Special case for the superuser. Useless unless we can
	    # somehow make this the top folder.
	    super = self._super	
	    if self._isTop() and (name==super.getUserName()):
	        return super

	    if cache.has_key(name):
		return cache[name]

	    data = self.getUserNames()
	    if name in data:
		user = User(name,self)
		cache[name] = user
		user = user.__of__(self)
		return user
	    else:
		cache[name] = None
		return None

	finally:
	    _lock[gufid].release()

    def _stashhash(self,username,passwd,request):
	'''stash the hash in our _hashmap cache. Set the cookie.'''

	gufid = self._gufid

	response = request['RESPONSE']
	hash = encodestring(sha.new('%s%s%f%f%s' \
	    % (username,passwd,time(),whrandom.random(),str(request))).digest())

	hashmap = _hashmap[gufid]

	# The Universe will probably end before a sha hash isn't unique,
	# but call me paranoid
	while hashmap.has_key(hash):
	    hash = encode_string(decode_string(hash) \
		+ chr(whrandom.randint(ord('A'),ord('Z'))))

	hashmap[hash] = [username,time()]# List rather than tuple so its mutable

	# TODO: Calculate path so that cookie only gets passed to the
	# acl_users folder's parent folder and below?
	response.setCookie('_gufauth',quote(hash),path='/')

    def _recurse_validate(self,request,auth,roles,\
		username=None,passwd=None,hash=None):
	'''Try validating locally, or crawl up the tree until
	   we succeed or reach the top'''

	gufid = self._gufid

	# Check from BasicUserFolder.validate. I don't actually know
	# what it does except that in some circumstances it will deny
	# access (better safe than sorry)
	if roles is _what_not_even_god_should_do:
	    _debug('_what_not_even_god_should_do request')
	    request.response.notFoundError()

	parents = []
	if request.has_key('PARENTS'):
	    parents=request['PARENTS']
	    parent=parents[0]
	else:
	    parent=self.aq_parent

	# If we have a password, then someone has just submitted the login
	# form
	if passwd is not None:
	    user = self.getUser(username)
	    if user is not None and user.authenticate(passwd,request) \
		    and user.allowed(parent,roles):
		self._stashhash(username,passwd,request)
		_debug('Authenticated %s using _guf_password field' \
		    % user.getUserName())
		return user

	# If we have a hash, then we are confirming the validity of a cookie
	if hash is not None:

	    hashmap = _hashmap[gufid]
	    _lock[gufid].acquire()
	    try:
		    
		# If we are using cookie authentication, check our hashmap
		# stash. We do it this way so that our hashes are shared
		# between threads (otherwise, you would have to authenticate
		# once for each thread that handles one of your requests).
		if hashmap.has_key(hash):
		    if time() > hashmap[hash][1] + self.cookie_expires:
			# Our cookie has expired!
			del hashmap[hash]
		    else:
			# Update our lastseen time
			hashmap[hash][1] = time()
			username = hashmap[hash][0]

			user = self.getUser(username)
			if user is not None and user.allowed(parent,roles):
			    _debug('Cookie authorized for %s' \
				% user.getUserName())
			    return user
	    finally:
		_lock[gufid].release()

	# We have failed to validate here. We now need to climb the
	# tree and try validating with other acl_users folders.
	# Although Zope also does this, we can't use it: The only way
	# of displaying a nice HTML login screen is by throwing an exception
	# out of the validate method, which has the side effect of causing
	# 'higher' acl_users folders to not be checked.
	#
	# Other UserFolder products that handle cookies (etcUserFolder and
	# UserDB) will just return 'None' from this function if auth is
	# not none - however, this doesn't work with REMOTE_USER 
	# authentication.

	return None

	if auth is not None and auth != '':
	    _debug('Auth set - returning None from _recurse_validate')
	    return None

	lastuserfolder = self
	if hasattr(parent,'aq_parent'):
	    c = parent.aq_parent
	    while c is not None:
		if hasattr(c,'aq_explicit') \
		    and hasattr(c.aq_explicit,'acl_users'):
		    
		    u = c.aq_explicit.acl_users

		    # This check is probably just paranoia
		    if u.absolute_url() != lastuserfolder.absolute_url():
			#_debug("Found new acl_users: %s" % u.absolute_url())
			lastuserfolder = u

			if hasattr(u,'isAGenericUserFolder'):
			    # If we find a GUF above, we won't need to
			    # crawl up any further as its validate method
			    # will do it for us
			    user = u.validate(request,auth,roles)
			    if user:
				return user
			    else:
				return None

			if hasattr(u,'isAUserFolder') and u.isAUserFolder:
			    user = u.validate(request,auth,roles)
			    if user:
				return user

		if hasattr(c,'aq_parent'):
		    c = c.aq_parent
		else:
		    c = None
	return None


    def _cookie_validate(self,request,auth,roles=None):
        response=request['RESPONSE']

	cname = '_gufauth' 
	nameparm = cname + '_name'
	passparm = cname + '_password'

	username = None
	password = None
	hash = None

	gufid = self._gufid

	if request.form.has_key(nameparm):
	    #
	    # Pull username & cleartext from form input.
	    # Ensure cleartext isn't left in the REQUEST for
	    # malicious DTML
	    #
	    username = request[nameparm]
	    del request.form[nameparm]
	    request[nameparm] = 'This space intentionally blank'

	if request.form.has_key(passparm):
	    password = request[passparm]
	    del request.form[passparm]
	    request[passparm] = 'This space intentionally blank'
	
	e = request.environ
	if _remote_user_mode and e.has_key('REMOTE_USER'):
	    username = e['REMOTE_USER']
	    _debug('REMOTE_USER mode detected - forcing username to %s'
		% username)
    
	if request.cookies.has_key(cname):
	    # Pull username & hash from cookie	
	    hash = unquote(request.cookies[cname])  

	if username is not None and password is not None:
	    #
	    # Try validating with username & password, or fall through to
	    # using 'auth' on our parent folders.
	    #
	    user = self._recurse_validate(request,auth,roles,username,password)

	    if user is not None:
		# If user is relogging in again, nuke their
		# old cookie right now
		if hash is not None:
		    try: del _hashmap[gufid][hash]
		    except KeyError: pass
		return user

	if hash is not None:
	    user = self._recurse_validate(request,auth,roles,hash=hash)

	    if user: 
		return user
	    else:
		response.expireCookie('_gufauth', path='/')

	if auth is not None and auth != '':
	    return None

	# We can log a message here, as we already have tried to authenticate
	# against 'higher' acl_users folders. A nice side effect of having
	# to do this work ourselves. This log message might mean a bad
	# password, or a valid user trying to access something they have
	# no rights to see.
	if username is not None and username != '':
	    LOG('GUF',0,'Failed authorization for user %s (%s)' \
		% (username,self.absolute_url()))

	raise 'LoginRequired', self.docLogin(self, request)
	

    def validate(self,request,auth='',roles=None):
	# No docstring for a reason

	_debug('Validate auth=%s,roles=%s' % (str(auth),str(roles)))

	# Two days later, I finally get a work around. If I get None
	# passed as the auth information, fake a Basic Authentication
	# auth string for no username and no password. I still don't
	# understand why this is needed or why it is working, but I have
	# other projects to get going on...
	#
	# If you are feeling brave, comment out the next two lines.
	# Then authentication to a GUF section of your web site and try
	# and access object/manage_workspace.
	#
	#if auth is None:
	#    auth = 'Basic Og=='

	if auth is None:
	    auth = ''

        if self.cookie_mode:
            u = self._cookie_validate(request, auth, roles)
        else:
	    u = BasicUserFolder.validate(self,request, auth, roles)
	return u
	
    def __len__(self):
        return len(self.getUserNames())

    def manage_edit(self, cookie_mode, userlist_expires,user_expires,
	    cookie_expires, REQUEST=None):
	"""Change properties"""

        self.cookie_mode=cookie_mode
	self.userlist_expires=userlist_expires
	self.user_expires=user_expires
	self.cookie_expires=cookie_expires

        if REQUEST:
            return MessageDialog(
                   title  ='GenericUserFolder Changed',
                   message='GenericUserFolder properties have been updated',
                   action =REQUEST['URL1']+'/manage_properties',
                   target ='manage_main')

    # Multiple inheritance wierdness
    #
    def manage_beforeDelete(self,item,container):
	Folder.manage_beforeDelete(self,item,container)
	BasicUserFolder.manage_beforeDelete(self,item,container)

    def manage_afterAdd(self,item,container):
	Folder.manage_afterAdd(self,item,container)
	BasicUserFolder.manage_afterAdd(self,item,container)

    def _killcookie(self,request):
	'Helper method to login and logoff'
	if not self.cookie_mode: raise 'Method only works in cookie mode'
	cname='_gufauth'
	if request.cookies.has_key(cname):
	    # Pull username & hash from cookie	
	    hash = unquote(request.cookies[cname])  
	    del _hashmap[self._gufid][hash]
	request['RESPONSE'].expireCookie(cname,path='/')

    def login(self, REQUEST):
	"""Login"""
	self._killcookie(REQUEST)
        return self.docLogin(self, REQUEST)

    def logout(self, REQUEST):
        """Logout"""
	self._killcookie(REQUEST)
        return self.docLogout(self, REQUEST)

    def _templateFile(self,file,title='',template='dumb'):
	t = HTMLFile('templates/%s/%s' % (template,file),globals())
	self.manage_addDTMLMethod(file,title,t)
    
    def _templateReset(self, template='dumb'):
	self._templateFile('userList','List Users',template)
	self._templateFile('userRoles',"Return a user's roles",template)
	self._templateFile('userDomains','Return domains a user can login from',
	    template)
	self._templateFile('userAuthenticate','Authenticate a user',template)
	self._templateFile('docLogin','Cookie mode login screen',template)
	self._templateFile('docLogout','Cookie mode logout screen',template)

manage_addGenericUserFolderForm=HTMLFile('addGenericUserFolder', globals())

def manage_addGenericUserFolder(self,cookie_mode=0,
				template='dumb',REQUEST=None):
    'Create a GenericUserFolder object'

    if hasattr(self.aq_base, 'acl_users'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a User Folder',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])

    ob=GenericUserFolder(cookie_mode, REQUEST)
    self._setObject('acl_users', ob)
    self.__allow_groups__=self.acl_users

    ob._templateReset(template)

    if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)

